; 
;
; Instituto Superior Tecnico
; Licenciatura de Engenharia Informatica e de Computadores
;
; Projecto de Fundamentos de Programacao 2005/2006
;
; no 56907 - Andre dos Santos Matias
; no 57006 - Pedro Miguel da Rocha Pinto

; TAI Tensor

;;;;;;;;;;;;;;
; CONSTRUTORES
;;;;;;;;;;;;;;

; v : universalk -> vector
;
; v(obj1,...,objk) devolve um vector com k elementos, preenchido com os elementos obj1...objk
;
(define v list)

; cria-vector : inteiro x universal -> vector
;
; cria-vector(k,obj) devolve um vector com as dimensoes referidas no inteiro k, preenchido com o
;                    elemento obj
;
(define (cria-vector k obj)
  (if (menor k 1)
      ()
      (cons obj (cria-vector (subtrai k 1) obj))))

; insere-vector : universal x vector -> vector
;
; insere-vector(obj,k,v) devolve um vector que resulta de inserir o elemento obj na primeira
;                        posicao do vector v.
;
(define (insere-vector obj v)
  (cons obj v))

; cria-tensor : vector x universal -> tensor
;
; cria-tensor(v,obj) devolve um tensor com as dimensões referidas no vector v, preenchido com o 
;                    elemento obj
;
(define (cria-tensor v obj)
  (define (cria-tensor-aux v obj)
    (cond ((null? v) obj)
          ((null? (cdr v)) (cria-vector (car v) obj))
          ((menor (car v) 1) ())
          (else (cons (cria-tensor (cdr v) obj)
                      (cria-tensor (cons (subtrai (car v) 1)
                                         (cdr v)) obj)))))
  (if (vector? v)
      (cria-tensor-aux v obj)
      (error "cria-tensor: o primeiro argumento nao e um vector")))

; junta-tensor : tensor x tensor -> tensor
;
; junta-tensor(t1,t2) no caso de os dois argumentos serem escalares, devolve um vector contendo
;                     contendo esses argumentos, no caso de os dois argumentos serem tensores,
;                     devolve um tensor que junta o tensor t1 e tensor t2 ao longo da sua ultima
;                     dimensao
;
(define (junta-tensor t1 t2)
  (cond ((escalar? t1) (v t1 t2))
        ((null? t1) t2)
        ((vector? t1) (cons (car t1)
                            (junta-tensor (cdr t1) t2)))
        ((tensor=? (forma-tensor t1) (forma-tensor t2)) (map junta-tensor t1 t2))
        (else (error "junta-tensor: os argumentos devem ter o mesmo tamanho"))))

;;;;;;;;;;;;
; SELECTORES
;;;;;;;;;;;;

; acede-tensor : tensor x inteiro -> elemento
;
; acede-tensor(t,k) tem como valor o elemento na posicao k da primeira dimensao do tensor t. se o
;                   tensor t é um escalar, ou k for inferior a um ou superior ao numero de elementos
;                   da primeira dimensao do tensor t, o valor da operacao e indefinido
;
(define (acede-tensor t k)
  (cond ((null? t) (error "acede-tensor: posicao fora da lista"))
        ((igual k 1) (car t))
        (else (acede-tensor (cdr t) (subtrai k 1)))))

; resto-tensor : tensor x inteiro -> tensor
;
; resto-tensor(t,k) tem como valor o tensor que resulta de remover o elemento que se encontra na
;                   posicao k da primeira dimensao do tensor t. se o tensor t é um escalar, ou k
;                   for inferior a um ou superior ao numero de elementos da primeira dimensao do
;                   tensor t, o valor da operacao e indefinido
;
(define (resto-tensor t k)
  (cond ((null? t) (error "resto-tensor: posicao fora do tensor"))
        ((igual k 1) (cdr t))
        (else (cons (car t)
                    (resto-tensor (cdr t) (subtrai k 1))))))

; remove-tensor : tensor x inteiro -> tensor
;
; remove-tensor(t,k) devolve um tensor onde foram removidos k elementos do inicio da primeira dimensao
;                    (se k > 0), ou fim (se k < 0)
;
(define (remove-tensor t n)
  (let ((comp (length t)))
    (cond ((and (maior n 0) (menor-igual n comp))
           (list-tail t n))
          ((and (menor n 0) (maior-igual n (subtrai comp)))
           (reverse (list-tail (reverse t) (multiplica n -1))))
          (else (error "remove-tensor: posicao fora do tensor")))))

; comp-tensor : tensor -> inteiro
;
; comp-tensor(t) tem como valor o numero de elementos na primeira dimensao do tensor t. se o tensor t
;                for um é escalar o valor desta operacao e indefinido
;
(define (comp-tensor t)
  (if (null? t)
      0
      (soma 1 (comp-tensor (cdr t)))))

; forma-tensor : tensor -> vector
;
; forma-tensor(t) tem como valor um vector cujos elementos correspondem ao comprimento de cada
;                 dimensao do tensor t. se o tensor t for um escalar, devolve um vector vazio
;
(define (forma-tensor t)
    (cond ((number? t) ())
          ((vector? t) (v (comp-tensor t)))
          (else (junta-tensor (v (comp-tensor t)) (forma-tensor (car t))))))

;;;;;;;;;;;;;;;
; MODIFICADORES
;;;;;;;;;;;;;;;

; altera-tensor! : tensor x tensor -> tensor
;
; altera-tensor!(t1,t2) muda os elementos do tensor t1 e devolve o tensor t1. os elementos sao
;                       obtidos atraves da enumeracao dos elementos do tensor t2 segundo argumento,
;                       repetindo essa enumeracao as vezes que forem necessarias para preencher o
;                       tensor t1
;
(define (altera-tensor! t1 t2)
  (let* ((vect (tensor->vector t2))
         (comp (comp-tensor vect))
         (i 0))
    (mapa-tensor (lambda (x)
                   (if (igual i comp)
                       (set! i 0))
                   (set! i (soma i 1))
                   (acede-tensor vect i))
                 t1)))

;;;;;;;;;;;;;;;;
; RECONHECEDORES
;;;;;;;;;;;;;;;;

; escalar? : universal -> logico
;
; escalar?(obj) tem o valor verdadeiro, se obj é um escalar, e tem o valor falso caso contrario
;
(define escalar? number?)

; vector? : universal -> logico
;
; vector?(obj) tem o valor verdadeiro, se obj é um vector, e tem o valor falso caso contrario
;
(define (vector? obj)
  (or (null? obj)
      (and (list? obj)
           (number? (car obj))
           (vector? (cdr obj)))))

; vector-vazio? : vector -> logico
;
; vector-vazio?(v) tem o valor verdadeiro, se o vector v é vector vazio, e tem o valor falso caso
;                  contrario
;
(define vector-vazio? null?)

; matriz? : universal -> logico
;
; matriz?(obj) tem o valor verdadeiro, se obj é uma matriz, e tem o valor falso caso contrario
;
(define (matriz? obj)
  (define (matriz?-aux obj forma)
    (if (not (null? obj))
        (and (vector? (car obj))
             (equal? forma (forma-tensor (car obj)))
             (matriz?-aux (cdr obj) forma))
        #t))
  (if (and (list? obj)
           (not (null? obj))
           (vector? (car obj)))
      (matriz?-aux (cdr obj) (forma-tensor (car obj)))
      #f))

; tensor? : universal -> logico
;
; tensor?(obj) tem o valor verdadeiro, se o obj é um tensor, e tem o valor falso caso contrario
;
(define (tensor? obj)
  (define (tensor?-aux obj forma)
    (if (not (null? obj))
        (and (list (car obj))
             (equal? forma (forma-tensor (car obj)))
             (tensor?-aux (cdr obj) forma))
        #t))
  (or (number? obj)
      (vector? obj)
      (and (list? obj)
           (if (vector? (car obj))
               (matriz? obj)
               (and (tensor? (car obj))
                    (tensor?-aux (cdr obj) (forma-tensor (car obj))))))))

;;;;;;;;
; TESTES
;;;;;;;;

; tensor=? : tensor x tensor -> logico
;
; tensor=?(t1,t2) tem o valor verdadeiro se o tensor t1 é igual ao tensor t2, e tem valor falso
;                 caso contrario
;
(define (tensor=? t1 t2)
  (and (equal? (forma-tensor t1) (forma-tensor t2))
       (not (membro-vector? 0 (tensor->vector (mapa-tensor (logico->escalar equal?) t1 t2))))))

; membro-vector? : universal x vector -> logico
;
; membro-vector?(obj,v) tem o valor verdadeiro se o obj existe no vector, e tem valor falso caso
;                       contrario
;
(define (membro-vector? obj v)
  (if (membro obj v)
     #t
     #f))

;;;;;;;;;;;;;;;;;
; TRANSFORMADORES
;;;;;;;;;;;;;;;;;

; mapa-tensor : procedimento x tensork -> tensor
;
; mapa-tensor(proc,t1,...,tk) devolve um tensor com as mesmas dimensoes dos seus argumentos, se
;                             for dado apenas um procedimento e um tensor como argumento, aplica
;                             o procedimento a cada um dos elementos da ultima dimensao do 
;                             tensor, se for dado mais que um tensor como argumento,  aplica o
;                             procedimento a cada elemento das ultimas dimensoes dos tensores
;                             na mesma posicao. a ordem em que o procedimento é aplicado aos 
;                             elementos não é especificada. se os tensores argumento nao tiverem
;                             as mesmas dimensoes, o resultado e indefinido
;
(define (mapa-tensor proc . t)
  (define (mapa-tensor-aux t)
    (cond ((or (null? t) (null? (car t))) ())
          ((vector? t) (apply proc t))
          (else (cons (mapa-tensor-aux (map car t))
                      (mapa-tensor-aux (map cdr t))))))
  (cond ((or (number? t) (number? (car t))) (apply proc t))
        ((null? (car t)) ())
        ((vector? (car t)) (cons (apply proc (map car t))
                                 (apply mapa-tensor (cons proc 
                                                          (map cdr t)))))
        (else (mapa-tensor-aux t))))

; mapa-dim : procedimento x tensork -> tensor
;
; mapa-dim(proc,t1,...,tk) devolve um tensor com as mesmas dimensoes dos seus argumentos, se
;                          for dado apenas um procedimento e um tensor como argumento, aplica
;                          o procedimento a cada um dos elementos da primeira dimensao do 
;                          tensor, se for dado mais que um tensor como argumento,  aplica o
;                          procedimento a cada elemento das primeiras dimensoes dos tensores
;                          na mesma posicao. a ordem em que o procedimento é aplicado aos 
;                          elementos não é especificada. se os tensores argumento nao tiverem
;                          as mesmas dimensoes, o resultado e indefinido
;
(define mapa-dim map)

; funcao-tensor : procedimento x tensor x tensor -> tensor
;
; funcao-tensor(proc,t1,t2) devolve um tensor que resulta de aplicar o procedimento aos elementos dos
;                           tensores argumentos. se os argumentos forem tensores do mesmo tamanho e 
;                           forma, o tensor resultado tem o mesmo tamanho e forma dos argumentos e tem,
;                           como elementos, o resultado de aplicar o procedimento aos elementos 
;                           correspondentes dos tensores argumentos. se um dos argumentos for um ;                           escalar, o tensor resultado tem o mesmo tamanho e forma que o outro e tem, 
;                           como elementos, o resultado de aplicar o procedimento do argumento escalar 
;                           com os elementos do outro argumento. em qualquer outro caso, o resultado e
;                           um erro
;
(define (funcao-tensor proc t1 t2)
  (cond ((escalar? t1) (mapa-tensor (lambda (x) (proc t1 x)) t2))
        ((escalar? t2) (mapa-tensor (lambda (x) (proc x t2)) t1))
        ((tensor=? (forma-tensor t1) (forma-tensor t2)) (mapa-tensor proc t1 t2))
        (else (error "funcao-tensor: os argumentos nao tem a mesma forma"))))


; escreve-tensor : tensor -> void
;
; escreve-tensor(t) escreve o tensor t
;
(define (escreve-tensor t)
  (define (escreve-vector t)
    (define (escreve-vector-aux t)
      (if (not (null? (cdr t)))
          (begin (display (car t))
                 (display " ")
                 (escreve-vector-aux (cdr t)))
          (display (car t))))
    (if (not (null? t))
        (escreve-vector-aux t)))
  (define (escreve-tensor-aux t s)
    (cond ((number? t) (display t))
          ((vector? t) (escreve-vector t))
          (else (if (not (null? (cdr t)))
                    (if (not (null? (cdr s)))
                        (begin (escreve-tensor-aux (car t) (cdr s))
                               (escreve-n (subtrai (comp-tensor s) 2))
                               (escreve-tensor-aux (cdr t) s)))
                    (escreve-tensor-aux (car t) (cdr s))))))
  (define (escreve-n n)
    (if (or (igual n 0) (menor n 0))
        (newline)
        (begin (newline)
               (escreve-n (subtrai n 1)))))
  (escreve-tensor-aux t (forma-tensor t)))

; tensor->vector : tensor -> vector
;
; tensor->vector(t) tranforma o tensor t num vector
;
(define (tensor->vector t)
  (cond ((number? t) (v t))
        ((vector? t) t)
        (else (junta-tensor (tensor->vector (car t))
                            (tensor->vector (cdr t))))))

(define igual =)
(define soma +)
(define subtrai -)
(define multiplica *)
(define divide /)
(define div-inteira quotient)
(define div-resto remainder)
(define raizq sqrt)
(define seno sin)
(define coseno cos)
(define menor <)
(define maior >)
(define menor-igual <=)
(define maior-igual >=)
(define membro member)

; logico->escalar
;
(define (logico->escalar proc)
  (lambda (x y)
    (if (proc x y)
        1
        0)))

;; Funções Monádicas

; display-tensor
;
(define (display-tensor t)
  (escreve-tensor t))
  

; symmetrical
;
(define (symmetrical t)
  (mapa-tensor (lambda (x) (multiplica x -1)) t))

; inverse
;
(define (inverse t)
  (mapa-tensor (lambda (x) (expt x -1)) t))

; !
;
(define (! t)
  (letrec ((factorial (lambda (n)
                     (if (igual n 0)
                         1
                         (multiplica n (factorial (subtrai n 1)))))))
    (mapa-tensor factorial t)))

; sqrt
;
(define (sqrt t)
  (mapa-tensor raizq t))

; sin
;
(define (sin t)
  (mapa-tensor seno t))

; cos
;
(define (cos t)
  (mapa-tensor coseno t))

; ~
;
(define (~ t)
  (let ((negacao (lambda (n)
                   (if (igual n 0)
                       1
                       0))))
    (mapa-tensor negacao t)))

; shape
;
(define (shape t)
  (forma-tensor t))

; interval
;
(define (interval n)
  (define (interval-aux resto acc)
    (if (igual resto 0)
        (v)
        (insere-vector acc (interval-aux (subtrai resto 1) (soma acc 1)))))
  (interval-aux n 1))

;; Funções Diádicas

; +
;
(define (+ t1 t2)
  (funcao-tensor soma t1 t2))

; -
;
(define (- t1 t2)
  (funcao-tensor subtrai t1 t2))

; *
;
(define (* t1 t2)
  (funcao-tensor multiplica t1 t2))

; /
;
(define (/ t1 t2)
  (funcao-tensor divide t1 t2))

; quotient
;
(define (quotient t1 t2)
  (funcao-tensor div-inteira t1 t2))

; reminder
;
(define (remainder t1 t2)
  (funcao-tensor div-resto t1 t2))

; <
;
(define (< t1 t2)
  (funcao-tensor (logico->escalar menor) t1 t2))

; >
;
(define (> t1 t2)
  (funcao-tensor (logico->escalar maior) t1 t2))

; <=
;
(define (<= t1 t2)
  (funcao-tensor (logico->escalar menor-igual) t1 t2))

; >=
;
(define (>= t1 t2)
  (funcao-tensor (logico->escalar maior-igual) t1 t2))

; =
;
(define (= t1 t2)
  (funcao-tensor (logico->escalar igual) t1 t2))

; ||
;
(define (|| t1 t2)
  (funcao-tensor (logico->escalar (lambda (x y) (not (and (zero? x) (zero? y))))) t1 t2))

; &&
;
(define (&& t1 t2)
  (funcao-tensor (logico->escalar (lambda (x y) (not (or (zero? x) (zero? y))))) t1 t2))

; drop
;
(define (drop t1 t2)
  (define (drop-aux t1 t2)
    (cond ((escalar? t1) (remove-tensor t2 t1))
          ((vector-vazio? (resto-tensor t1 1)) (remove-tensor t2 (acede-tensor t1 1)))
          (else (mapa-dim (lambda (x)
                            (drop-aux (resto-tensor t1 1) x))
                          (remove-tensor t2 (acede-tensor t1 1))))))
  (if (escalar? t2)
      (error "drop: o segundo argumento nao pode ser um escalar")
      (drop-aux t1 t2)))
; reshape
;
(define (reshape v t)
  (let ((t-final (cria-tensor v 0))
        (t-aux (tensor->vector t)))
    (altera-tensor! t-final t-aux)))

; catenate
;
(define (catenate t1 t2)
  (junta-tensor t1 t2))

; member
;
(define (member t1 t2)
  (let ((v (tensor->vector t2)))
    (mapa-tensor (lambda (x)
                   ((logico->escalar membro-vector?) x v))
                 t1)))

; select
;
(define (select v t)
  (letrec ((i 0)
           (filtro-vector
            (lambda (proc? v1)
              (cond ((vector-vazio? v1) v1)
                    ((proc? (acede-tensor v1 1))
                     (insere-vector (acede-tensor v1 1)
                                    (filtro-vector proc? (resto-tensor v1 1))))
                    (else (filtro-vector proc? (resto-tensor v1 1))))))
           (testa?
            (lambda (x)
              (if (igual i (comp-tensor v))
                  (set! i 0))
              (set! i (soma i 1))
              (if (igual 0 (acede-tensor v i))
                  #f
                  #t))))    
    (cond ((escalar? t) ())
          ((vector? t) (filtro-vector testa? t))
          (else (mapa-dim (lambda (x)
                       (select v x))
                     t)))))

;; Operadores Monádicos

; fold
;
(define (fold proc)
  (let* ((fold-aux (lambda (x)
                     (if (vector-vazio? (resto-tensor x 1))
                         (acede-tensor x 1)
                         (proc (acede-tensor x 1) ((fold proc) (resto-tensor x 1)))))))
    (lambda (x)
      (if (vector-vazio? x)
          (v)
          (fold-aux x)))))

; scan
;
(define (scan proc)
  (lambda (x)
    (if (vector-vazio? x)
        (v)
        (junta-tensor ((scan proc) (resto-tensor x (comp-tensor x))) 
                      (v ((fold proc) x))))))

; outer-product
;
(define (outer-product proc)
  (lambda (t1 t2)
    (mapa-tensor (lambda (x)
           (mapa-tensor (lambda (y)
                  (proc x y))
                t2))
         t1)))

;; Operadores Diadicos

; inner-product
;
(define (inner-product proc1 proc2)
  (define (transpose m)
    (cond ((not (matriz? m)) (v))
          ((vector-vazio? (resto-tensor (acede-tensor m 1) 1))
           (insere-vector (mapa-dim (lambda (x) (acede-tensor x 1)) m)
                          (v)))
          (else (insere-vector (mapa-dim (lambda (x) (acede-tensor x 1)) m)
                               (transpose (mapa-dim (lambda (x) (resto-tensor x 1)) m))))))
  (lambda (x y)
    (cond ((or (escalar? x) (escalar? y)) (proc2 x y))
          ((and (vector? x)
                (vector? y)
                (tensor=? (shape x) (shape y)))
           (if (not (vector-vazio? x))
               ((fold proc1) (mapa-tensor proc2 x y))))
          (else (mapa-dim (lambda (x)
                            (mapa-dim (lambda (y)
                                        ((inner-product proc1 proc2) x y)) (transpose y))) x)))))

;; 4 Exercicios

; 1 tally
;
; dado um tensor, devolve o numero de elementos desse tensor
;
(define (tally t)
  ((fold *) (shape (v t))))

; 2 rank
;
; dado um tensor, devolve um escalar com o numero de dimensoes do tensor
;
(define (rank t)
  (reshape () (shape (shape t))))

; 3 average
;
; dado um vector de numeros, devolve um escalar representado a media dos numeros do vector
;
(define (average v)
  (/ ((fold +) v)  (reshape () (shape v))))

; 4 within 
;
; dado um vector de números e dois números, devolve um vector que apenas contém os elementos
; do vector cujo valor é maior ou igual a n1 e menor ou igual a n2
(define (within v n1 n2)
  (let ((v2 (select (<= v n2) v)))
    (select (>= v2 n1) v2)))

; 5 substitute<
;
; dado um vector de numeros e dois numeros, devolve um vector em que todos os elementos do vector
; que sejam inferiores ao primeiro numero sao substituidos pelo segundo numero
;
(define (substitute< v n1 n2)
  (+ (* (>= v n1) v) (* (< v n1) n2)))

; 6 ravel
;
; dado um tensor, devolve um vector contendo todos os elementos do tensor
;
(define (ravel t)
  (reshape (v ((fold *) (shape (v t)))) t))

; 7 primes
;
; dado um escalar, devolve um vector contendo todos os numeros primos desde 2 ate esse escalar,
; inclusive
;
(define (primes n)
  (let ((x (drop 1 (interval n))))
    (select (~ (member x ((outer-product *) x x))) x)))
